/**************************************************************
*
* Heated motorbike grips, each side independent controlled,
* one pushbutton operation.
*   click
*     switch on, more heat, set requested heat higher
*     LED blinks 1x
*   double click
*     stop heating, set requested heat lower
*     LED blinks 2x
*   long press more than 2 seconds
*     switch off
*     LED blinks 4x
*
* LED brigtness high if heat is on, else low.  
* 
* Hardware
* Arduino nano (other Arduino-boards should also work, check Pin types).
* pushbutton attached to Pin 8 and ground.
* LED attached from Pin 9 to ground (5V-type, others through appropriate resistor)
* PowerOut-Pin 10,11 (PWM default 490Hz if arduino nano is used) and ground. 
*   attach power transistors, MOSFETs, dual H-Bridge etc.
* 2 thermistors NTC 10K/25C   
*   thermistors attached to pins 0,1 and +5V, 
*   ground pin2 0,1 through 10K ohm resistors (voltage divider).
*   If 100KOhm thermistors are used, resistor value is 100KOhm.
* PowerOut Pin 10,11 to ground. max 40mA. 
* Connect to power stage (MosFET, H-Bridge etc. adequate for grip heaters demand, usually 2-3A).
* Power supply 5V on Pin "5V"or 6-15V on Pin Vin. 
* Caution, motor vehicles electric has short-time very high tensions and currents.
* Use appropriate circuitry (e.g. varistor, snubber).
* reset-PIN grounded through capacitor 10-100nF 
*
* libraries
*   PolymorphicButtons
*     https://github.com/JCWentzel/PolymorphicButtons/
*   PID
*     https://github.com/br3ttb/Arduino-PID-Library/
*
* developers: 
* This is a quick and dirty first approach.
* Should be properly done as state machine.
* improvement suggestions welcome!
* 
*
************************************************************** 
*
* KURZFASSUNG
* Ein-Knopf-Steuerung für Motorrad-Heizgriffe mit Arduino
* Klicken/Halten länger als 1 Sekunde: Einschalten, mehr Wärme
*   LED blinkt 1x.
* Doppelklick: weniger Wärme
*   LED blinkt 2x.
* gedrückt halten (länger als 4 Sekunden): Ausschalten
*   LED blinkt 4x.
*
* LED leuchtet hell, wenn geheizt wird, sonst schwach.
* 
* Taster an Pin 8 und Masse (Masse = GND)
* 2 Wärmefühler in oder an den Griffen oder den Heizpatronen
*   (NTC, Nennwiderstand 10KOhm bei 25 Grad Celsius), in Spannungsteiler: 
*   ein NTC an Pin0 und 5V, der andere an Pin1 und 5V,
*   an Pin0 und Pin1 jeweils ein Widerstand 10KOhm zu Masse.
*   Bei 100KOhm-Thermistoren 100KOhm.
* LED (für 5 Volt! Sonst passender Vorwiderstand) an Pin9 und Masse.
* Ausgang Heizgriff 1: Pin 10 und Masse, Ausgang Heizgriff 2: Pin 11 und Masse.
*   Die Ausgänge haben 5V 40mA (maximal eine LED). 
*   Damit werden über etwa 220Ohm-Widerstand Leistungsendstufen angesteuert (MosFET, H-Bridge usw.),
*   ausgelegt nach der maximalen Leistung der Heizung.
* Spannungsversorgung: 5V an Pin 5V und Masse, einige Arduinos erlauben an Pin Vin
*   6-20V.
* Die Versorgungsspannung für elektronische Schaltungen in KFZ muss sorgfältig   
* vorbereitet werden, da kann es kurzzeitig SEHR hohe Spannungen und Ströme geben!
* Stichworte: Drossel, Varistor, Snubber, TLE 4270-2
* Snubber: LC-Glied 220Ohm, 0,1 µF, Drossel 47 µH
* Direkt am Resetpin (und nicht 1 cm daneben!) Kondensator (10nF, 47nF oder 100nF) gegen Masse 
* 
* benötigte Bibliotheken 
*   PolymorphicButtons
*     https://github.com/JCWentzel/PolymorphicButtons/
*   PID
*     https://github.com/br3ttb/Arduino-PID-Library/
*
*****************************************************************/


///////// USER DEFINABLE VALUES /////////////////////////////////////////////////////////////////////////

// temperatures

/* 
temperature in Celsius to thermistor values (thermistor NTC 10/25C)
20C   465
25C   515 
30    568
37C   628
40C   750
45C   790
50C   850
*/
    
double grip1RequestedTemperature = 750;         // initial grip1 target temperature (628 ca. 37C)
double grip2RequestedTemperature = 750;         // initial grip2 target temperature (628 ca. 37C)
const byte temperatureIncrementStep = 10; 	    // incremental step temperature up/down (20 is about 2C)
byte triggerPoint = 4;                          // PID less aggressive settings applied when approaching requested temperature
const unsigned int maxAllowedTemperature = 970; 			
const unsigned int minAllowedTemperature = 445;       


///////// PIN ASSIGNMENTS /////////

const byte thermistorPin1 = 0;
const byte thermistorPin2 = 1;
const byte ledPin1 = 9;
const byte powerOutPin1 = 10;                        
const byte powerOutPin2 = 11;                        

/////// END USER DEFINABLE VALUES /////////////////////////////////////////////////////////////////



///////// TEMPERATURE CALCULATION /////////

double thermistor1Value;                            // raw value from analogRead
double thermistor2Value;                            // raw value from analogRead


// PID (proportional–integral–derivative controller)
#include "PID_v1.h"

double powerAmount1, powerAmount2;
// variables thermistorValue, requestedTemperature also have to be double

// aggressive and conservative tuning Parameters
// optimum: 10% over after rapid change
double aggKp=4, aggKi=0.25, aggKd=0.25;         // Kp proportional (speed), Ki integral (accuracy), 
double consKp=2, consKi=0.1, consKd=0.5;        // Kd derivative (damps oscillating)

PID myPID1(&thermistor1Value, &powerAmount1, &grip1RequestedTemperature, consKp, consKi, consKd, P_ON_M, DIRECT);
PID myPID2(&thermistor2Value, &powerAmount2, &grip2RequestedTemperature, consKp, consKi, consKd, P_ON_M, DIRECT);



//////// PUSHBUTTON EVENTS ////////

#include "PMButton.h"
PMButton button1(8);
byte button1State = 0;                              // 0 off, 1 more heat, 2 less heat



// set requested Temperature
void setRequestedTemperature(byte button1State)
{
  if (button1State == 1)                                        // more heat                   
  {
    if (grip1RequestedTemperature < maxAllowedTemperature )
      grip1RequestedTemperature = grip1RequestedTemperature + temperatureIncrementStep;
    if (grip2RequestedTemperature < maxAllowedTemperature )
      grip2RequestedTemperature = grip2RequestedTemperature + temperatureIncrementStep;     
  }
  
  if (button1State == 2)                                         // less heat
  {
    if (grip1RequestedTemperature > minAllowedTemperature )
      grip1RequestedTemperature = grip1RequestedTemperature - temperatureIncrementStep;
    if (grip2RequestedTemperature > minAllowedTemperature )
      grip2RequestedTemperature = grip2RequestedTemperature - temperatureIncrementStep; 
  }  
}



// LED user feedback
// TODO replace delay with timer measurings
byte ledBlinks = 0;
void ledBlink(byte ledBlinks)
{
    if (ledBlinks == 1)                            
    {                                             
      if (powerAmount1 > 10 || powerAmount2 > 10)       // LED is already at max brigthness
      {
        analogWrite(ledPin1, 0);
        delay(200);    
      }

      else
      {
        analogWrite(ledPin1, 125);
        delay(200);    
      }
    }

    if (ledBlinks == 2)                           
    {                                             
      if (powerAmount1 > 10 || powerAmount2 > 10)       // LED is already at max brigthness 
      {
        analogWrite(ledPin1, 0);
        delay(200);    
        analogWrite(ledPin1, 125);
        delay(200);    
        analogWrite(ledPin1, 0);
        delay(200); 
      }

      else
      {
        analogWrite(ledPin1, 125);
        delay(200);    
        analogWrite(ledPin1, 0);
        delay(200);            
        analogWrite(ledPin1, 125);
        delay(200);    
      }
  }
}



///////////////////////////////////////////////////////////////

void setup()
{
  
  //////// button timings (milliseconds) ////////
  // default values if not explicitly declared
  /*  
  // default values
  button1.debounce(20);                           
  button1.dcGap(200);                      // Time between clicks for Double click.
  button1.holdTime(1000);                  
  button1.longHoldTime(4000);              
  */

  button1.begin();


  //////// pin modes ////////

  pinMode(powerOutPin1, OUTPUT);
  pinMode(powerOutPin2, OUTPUT);
  pinMode(ledPin1, OUTPUT);


  // PID controller

  thermistor1Value = 1000;    
  thermistor2Value = 1000;
  
  myPID1.SetMode(AUTOMATIC);
  myPID2.SetMode(AUTOMATIC);
  myPID1.SetSampleTime(1000); 
  myPID2.SetSampleTime(1000); 


////////  DEBUG OUTPUT ////////
// don't use  in production

// Serial.begin(57600);   


} // end setup



///////////////////////////////////////////////////////////////

void loop()
{

  //////////////// TEMPERATURES //////////////// 
 
  if(analogRead(thermistorPin1) > 0)
  {
    thermistor1Value = analogRead(thermistorPin1);    
  }
  else 
    thermistor1Value = 1000;   
  if(analogRead(thermistorPin2) > 0)
  {
    thermistor2Value = analogRead(thermistorPin2);
  }
  else 
    thermistor2Value = 1000; 


  //////////////// BUTTON STATES ////////////////

  button1.checkSwitch();

  if(button1.held())
    {
    button1State = 0;                   // switch off
    ledBlink(2);  
    ledBlink(2);
    ledBlink(2);
    }

  if(button1.clicked())
  {                                     
      button1State = 1;   
      setRequestedTemperature(1);       // set requested temperatures higher    
      ledBlink(1);  
  }

  if(button1.doubleClicked())
  {                                     
    button1State = 2;
    setRequestedTemperature(2);         // set requested temperatures lower
    ledBlink(2); 
  }

  
  
  //////////////// POWER STATES ////////////////

  if (button1State == 0)                // switched off
  {
    digitalWrite(powerOutPin1, LOW);                 
    digitalWrite(powerOutPin2, LOW);
  }

  if (button1State > 0)                
  {  
    // PID controller 
    double gap1 = abs(grip1RequestedTemperature - thermistor1Value); 
    double gap2 = abs(grip2RequestedTemperature - thermistor2Value); 
    byte triggerPoint;

    if (gap1 < triggerPoint)
    {  
      myPID1.SetTunings(consKp, consKi, consKd);
    }
    if (gap1 >= triggerPoint)
    {
       myPID1.SetTunings(aggKp, aggKi, aggKd);
    }

    if (gap2 < triggerPoint)
    {  
      myPID2.SetTunings(consKp, consKi, consKd);
    }
    if (gap2 >= triggerPoint)
    {
       myPID2.SetTunings(aggKp, aggKi, aggKd);
    }

    myPID1.Compute();
    analogWrite(powerOutPin1, powerAmount1); 
    myPID2.Compute();
    analogWrite(powerOutPin2, powerAmount2);

  }



  //////// LED STATES ////////

  // LED change state only if both grip states are equal
  // e.g., one grip gets full power while the other gets reduced power 
  // because temperature is nearby target temperature,
  // LED shows full power state.

  if (button1State == 0)                              // switched off
    analogWrite(ledPin1, 0);

  if (button1State > 0)          
  {
    if (powerAmount2 > 10 || powerAmount2 > 10)       // full power
      analogWrite(ledPin1, 125);                      // ledBrightnessHigh
    else
      analogWrite(ledPin1, 10);                       // ledBrightnessLow
  }                      



  ////////  DEBUG OUTPUT ////////
  // don't use  in production
  
  /* 
  Serial.print("button state: ");
  Serial.println(button1State);  

  Serial.print("thermistor1 value: ");
  Serial.print(thermistor1Value);
  Serial.print("\t");
  Serial.print("thermistor2 value: ");
  Serial.println(thermistor2Value);

  Serial.print("grip1 requested value: ");
  Serial.print(grip1RequestedTemperature);
  Serial.print("\t");
  Serial.print("grip2 requested value: ");
  Serial.println(grip2RequestedTemperature);

  Serial.print("power1: ");
  Serial.print(powerAmount1);
  Serial.print("\t \t");
  Serial.print("power2: ");
  Serial.println(powerAmount2);

  Serial.println("------------------------------------------------------------------");
  */

} 
